
import datetime


def get_value(key, record):
    """Return value from different types of input records
    Accepts action record, or simple object
    """

    if record.get('@type', None) == 'action':
        record = record.get('object', None)

    value = record.get(key, None)

    return value

def get_instrument_record():
    """
    """
    record = {
        "@type": "WebAPI",
        "@id": "75d41a03-7763-4c8f-a65b-52f80cdc7c77",
        "name": "kraken_user",
        "description": "description"
    }

    return record

def get_action_record(object_record, result_record=None):
    """
    """

    action_record = {
        '@type': 'action',
        '@id': str(uuid.uuid4()),
        'name': 'kraken_user',
        'object': object_record,
        'instrument': get_instrument_record(),
        'startTime': datetime.datetime.now(),
        'actionStatus': 'activeActionStatus'
    }

    if result_record:
        action_record['endTime'] = datetime.datetime.now()
        action_record['actionStatus'] = 'completedActionStatus'
        action_record['result'] = result_record

    return action_record



    



import json
import datetime
import os

""" JSON
"""
def _default_json(obj):

    if isinstance(obj, dict):
        new_obj = {}
        for key, value in obj:
            new_obj[key] = _default_json(value)
    elif isinstance(obj, list):
        new_obj = []
        for item in obj:
            new_obj.append(_default_json(item))
    elif isinstance(obj, (datetime.date, datetime.datetime)):
        new_obj = obj.isoformat()
    else:
        new_obj = obj

  
    return new_obj

def _convert_string_to_dates(record):

    if isinstance(record, dict):
        new_record = {}
        for key, value in record.items():
            new_record[key] = _convert_string_to_dates(value)

    elif isinstance(record, list):
        new_record = []
        for item in record:
            new_record.append(_convert_string_to_dates(item))
    else:
        new_record = record
        try:
            if record:
                new_record = datetime.datetime.fromisoformat(record)
        except Exception as e:
            a=1
    return new_record


def dumps(record):
   
    return json.dumps(record, default=_default_json, indent=4)

def loads(string):
    record = json.loads(string)
    return _convert_string_to_dates(record)

def dump(filename, record):

    os.makedirs(os.path.dirname(filename), exist_ok=True)

    with open(filename,"w") as fw:
        json.dump(record, fw, default=_default_json, indent=4)


def load(filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename,"r") as fr:
        out_dict = json.load(fr, default=_default_json, indent=4)
    return _convert_string_to_dates(out_dict)



    



    
import copy
from kraken_user.helpers import json
import os
from kraken_user import kraken_user as m
from kraken_user.class_kraken_user import User
import pkg_resources


"""
Notes:
To access files in data directory, use:
new_path = pkg_resources.resource_filename('kraken_user', old_path)

"""

class Users:
    """
    Collection contains many objects

    Args:
        arg1 (str): The arg is used for...
        arg2 (str): The arg is used for...
        arg3 (str): The arg is used for...

    Attributes:
        record (dict): This is where we store attributes
        json (str): Record in json format
        
    """

    def __init__(self):
        self._records = []
        self._index = 0
        

    def __str__(self):
        """
        """
        return str(self._records)


    def __iter__(self):
        """Defines itself as an iterator
        """
        self._index=0
        return self

    def __next__(self):
        """
        """
        if self._index < len(self._records):
            item = self._records[self._index]
            self._index += 1
            return item
            
        else: 
            self._index = 0
            raise StopIteration

    
    def __repr__(self):
        """
        """
        return str(self._records)

    def __len__(self):
        return len(self._records)
    
    def __eq__(self, other):
        """
        """
        if type(self) != type(other):
            return False
            
        if self._records == other._records:
            return True
        return False
        
    def set(self, values):
        """
        """
        values = values if isinstance(values, list) else [values]
        for i in values:
            self._records.append(i)
        return True

    
    def get(self, property):
        """
        """
        return 

    
    def load(self, values):
        """
        """
        values = values if isinstance(values, list) else [values]
        for i in values:
            o = User()
            o.load(i)
            self._records.append(o)

        return True


    def dump(self): 
        """
        """
        records = []
        for i in self._records:
            records.append(i.dump())
        return records
        

    def set_json(self, value):
        """
        """
        record = json.loads(value)
        self.load(record)
        return True

    def get_json(self):
        """
        """
        return json.dumps(self.dump())

    @property
    def records(self):
        return self.dump()

    @records.setter
    def records(self, value):
        return self.load(value)
    
    @property
    def json(self):
        return self.get_json()

    @json.setter
    def json(self, value):
        return self.set_json(value)
        

    



    
import copy
from kraken_user.helpers import json
from kraken_user.helpers import things
import os
import pkg_resources
import requests
import uuid
import datetime
import hashlib
import copy

"""
Notes:
To access files in data directory, use:
new_path = pkg_resources.resource_filename('kraken_user', old_path)

"""

records = {}


"""
Cache
"""


def get_from_cache(membershipNumber):
    """
    """

    if not membershipNumber:
        return None
    
    if records.get(membershipNumber, None):
        return records.get(membershipNumber, None)

    return None


def post_to_cache(record):
    """
    """

    if not record:
        return None
    
    if isinstance(record, list):
        for i in record:
            post_to_cache(i)
        return
        
    records[record.get('membershipNumber', None)] = record

    return

def delete_from_cache(membershipNumber):
    """
    """
    if not membershipNumber:
        return None
        
    records.pop(membershipNumber, None)
    return 
    



"""
I/O
"""

def delete(record):
    """
    """
    url = 'https://data.krknapi.com/kraken_login'

    headers = {"Content-Type": "application/json", "Authorization":"bob"}
    data = json.dumps(record)
    r = requests.delete(url, data=data, headers=headers)
   
    delete_from_cache(record.get('membershipNumber', None))
    return


def post(record):
    """
    """
    
    post_to_cache(record)
    
    url = 'https://data.krknapi.com/kraken_login'

    headers = {"Content-Type": "application/json", "Authorization":"bob"}
    data = json.dumps(record)
    r = requests.post(url, data=data, headers=headers)

    return


def create_user(record):
    """Register a new user
    """


    program = record.get('programName', None)
    membershipNumber = record.get('membershipNumber', None)

    if not program:
        return False

    if not membershipNumber:
        return False

    # Check if record exists
    db_record = get_user(program, membershipNumber)
    if db_record:
        # Handle situation where user already exists
        return False

    original_record = copy.deepcopy(record)
    #
    record = add_event(record, 'create user', original_record)
    
    # Set active
    record = set_active_status(record)
    
    # Create record
    post(record)
    
    return record


def get_user(program_name, membershipNumber):
    """
    """

    if not program_name or not membershipNumber:
        return None
    
    if get_from_cache(membershipNumber):
        return get_from_cache(membershipNumber)
    

    url = 'https://data.krknapi.com/kraken_login'

    headers = {"Content-Type": "application/json", "Authorization":"bob"}

    params = {'@type': 'programMembership', 'programName': program_name, 'membershipNumber': membershipNumber}

    r = requests.get(url, params=params, headers=headers)


    if r.status_code != 200:
        return None
    
    record = json.loads(r.content)

    if isinstance(record, list) and len(record) > 0:
        return record[0]
    if isinstance(record, list) and len(record) == 0:
        return None

    post_to_cache(record)
    
    return record


def get_users(program_name):
    """
    """


    url = 'https://data.krknapi.com/kraken_login'

    headers = {"Content-Type": "application/json", "Authorization":"bob"}

    params = {'@type': 'programMembership', 'programName': program_name,}

    r = requests.get(url, params=params, headers=headers)

    records = json.loads(r.content)

    return records


"""
Credentials
"""


def add_credential(record, password_value, salt_value):
    """
    """

    hash_password = get_hash_password(password_value, salt_value)

    if not record.get('hasCredential', None):
        record['hasCredential'] = []

    # Cancel existing credentials
    for i in record.get('hasCredential', []):
        validTo =  i.get('validTo', None)
        if not validTo or validTo > datetime.datetime.now():
            i['validTo'] = datetime.datetime.now()

    # Create new credential
    credential = {
        "@type": "credential",
        "@id": str(uuid.uuid4()),
        "validFrom": datetime.datetime.now(),
        "validTo": None,
        "proof": {
              "type": "DataIntegrityProof",
              "verificationMethod": "hash",
              "proofPurpose": "assertionMethod",
              "proofValue": hash_password
        }
    }
    record['hasCredential'].append(credential)
    record = add_event(record, 'add credential', credential)
    return record


def authenticate_user(record, password_value, salt_value):
    """
    """

    hash_password = get_hash_password(password_value, salt_value)

    for i in record.get('hasCredential', []):
        proof = i.get('proof', {})
        c1 = i.get('validFrom', None) < datetime.datetime.now()
        c2a = i.get('validTo', None) == None
        c2b = i.get('validTo', None) and i.get('validTo', None) > datetime.datetime.now()
        c3 = proof.get('proofValue', None) == hash_password

        if c1 and (c2a or c2b) and c3:
            return True

    return False


def cancel_passwords(record):
    """Cancels all current passwords
    """

    for i in record.get('hasCredential', []):
        validTo =  i.get('validTo', None)
        if not validTo or validTo > datetime.datetime.now():
            i['validTo'] = datetime.datetime.now()

    record = add_event(record, 'remove credentials')
    
    return


"""
Roles
"""
def add_role(record, role):
    """
    """

    if not record.get('hasRole', None):
        record['hasRole'] = []

    # Create new credential
    role = {
        "@type": "role",
        "@id": str(uuid.uuid4()),
        "validFrom": datetime.datetime.now(),
        "validTo": None,
        "name": role
    }
    record['hasRole'].append(role)

    record = add_event(record, 'add role', role)

    return record

def remove_role(record, role):

    for i in record.get('hasRole', []):
        if i.get('name', None) == role:
            i['validTo'] = datetime.datetime.now()
            record = add_event(record, 'remove role', i)
    
    return record

def get_active_roles(record):

    active_roles = []
    for i in record.get('hasRole', []):
        c1 = i.get('validFrom', None) < datetime.datetime.now()
        c2a = i.get('validTo', None) == None
        c2b = i.get('validTo', None) and i.get('validTo', None) > datetime.datetime.now()

        if c1 and (c2a or c2b) :
            active_roles.append(i.get('name', None))
    return active_roles

def get_active_status(record):
    """Returns True if 'member' role is active
    """
    if 'member' in get_active_roles(record):
        return True
    return False

def set_active_status(record):
    """Add member role
    """
    if not get_active_status(record):
        record = add_role(record, 'member')
        record = add_event(record, 'set active')

    return record


def set_inactive_status(record):
    """Remove member role
    """

    record = remove_role(record, 'member')
    record = add_event(record, 'set inactive')
    return record

"""
Base record
"""


def get_base_record():
    """
    """

    record = {
        "@type": "programMembership",
        "@id": str(uuid.uuid4()),
        "hostingOrganization": None,
        "member": {
            "@type": "person",
            "@id": None,
            "email": None
        },
        "programName": None,
        "membershipNumber": None,
        "hasCredential": [],
        "hasRole": [],
        "event": []
    }
    
    return record

def get_hash_password(password_value, salt_value):
    """
    """
    m = hashlib.sha256()
    m.update(bytes(password_value,'UTF-8'))
    m.update(bytes(salt_value,'UTF-8'))
    m.digest()

    result = m.hexdigest()

    return result

"""
Events
"""


def add_event(record, name, object=None, status='completedActionStatus'):
    """
    """

    # Add event
    timeEnd = datetime.datetime.now() if status in['completedActionStatus', 'failedActionStatus'] else None
    creation_event = {
        "@type": "action",
        "@id": str(uuid.uuid4()),
        "name": name,
        "timeStart": datetime.datetime.now(),
        "timeEnd": timeEnd,
        "object": object,
        "actionStatus": status
    }

    if not record.get('event', None):
        record['event'] = []

    if not isinstance(record.get('event', None), list):
        record['event'] = [record.get('event', None)]
        
    record['event'].append(creation_event)

    return record




def add_login_event(record, remote_addr=None):
    """
    """

    # Add login event
    record = add_event(record, 'login')

    # Add session event
    record = add_event(record, 'session', None, 'activeActionStatus')

    return record

def add_failed_login_event(record, remote_addr=None):
    """
    """
    record = add_event(record, 'login', None, 'failedActionStatus')

    return record


def add_logout_event(record, remote_addr=None):
    """
    """
    # Add login event
    record = add_event(record, 'logout')

    # Update session event
    for i in record.get('event', []):
        status = i.get('actionStatus', None)
        name = i.get('name', None)
        timeEnd = i.get('timeEnd', None)
        if name == 'session' and status == 'activeActionStatus':
            i['timeEnd'] = datetime.datetime.now()
            i['actionStatus'] = 'completedActionStatus'
    return record

def is_logged_in(record):
    """
    """
    for i in record.get('event', []):
        status = i.get('actionStatus', None)
        name = i.get('name', None)
        timeEnd = i.get('timeEnd', None)
        if name=='session' and status == 'activeActionStatus':
            return True

    return False
    



"""
HTML templates
"""

def get_login_form(email=None, referrer=None, action_url = '/login'):
    """
    """
    content = f"""
    <form action='{{action_url}}' method='POST'>
    <input class='form-control' type='text' name='email' id='email' placeholder='email' value='{{email}}/>
    <input class="form-control" type='password' name='password' id='password' placeholder='password'/>
    <input type='hidden' name='referrer' id='referrer' value='{{referrer}}'/>
    <input class='btn btn-primary' type='submit' name='submit'/>
   </form>

    """
    return content

def get_register_form(email=None, referrer=None, action_url = '/register'):
    """
    """
    content = f"""
    <form action='{{action_url}}' method='POST'>
    <input class='form-control' type='text' name='email' id='email' placeholder='email' value='{{email}}/>
    <input class="form-control" type='password' name='password' id='password' placeholder='password'/>
    <input type='hidden' name='referrer' id='referrer' value='{{referrer}}'/>
    <input class='btn btn-primary' type='submit' name='submit'/>
   </form>

    """
    return content



    
import copy
import uuid
import os
import pkg_resources
from kraken_user.helpers import json
from kraken_user import kraken_user as m

"""
Notes:
To access files in data directory, use:
new_path = pkg_resources.resource_filename('kraken_user', old_path)

"""





class User:
    """
    The Vehicle object contains a lot of vehicles

    Args:
        arg1 (str): The arg is used for...
        arg2 (str): The arg is used for...
        arg3 (str): The arg is used for...

    Attributes:
        _objects: List of all objects initialized with this class
        record (dict): This is where we store attributes
        json (str): Record in json format
        
    """

    _objects = []    # Initialize array to store all objects instances

    def __init__(self, program=None, membershipNumber=None):
        
        User._objects.append(self)    # Add itself to list of objects instances

        self._id = str(uuid.uuid4())
        self._record = m.get_base_record()
        self._index = 0
        self._salt = '36b8e053-2c43-4107-b51f-e2ebd05f3838'

        self._in_db = False

        if program:
            self.set_programName(program)

        if membershipNumber:
            self.set_membershipNumber(membershipNumber)

    
    def __str__(self):
        """
        """
        return str(self._record)

    
    def __repr__(self):
        """
        """
        return str(self._record)

    def __iter__(self):
        """ Allows the object to be used in a for loop as a array of only itself
        """
        self._index=0
        return self

    def __next__(self):
        if self._index == 0:
            self._index = 1
            return self
        else:
            self._index = 0
            raise StopIteration


    
    def __eq__(self, other):
        """
        """
        if type(self) != type(other):
            return False
            
        if self._record == other._record:
            return True
        return False

    def __gt__(self, other):
        """
        """
        return True



    @property
    def id(self):
        return self.get_id()

    @id.setter
    def id(self, value):
        self.set_membershipNumber(value)
    
        
    def set(self, property, value):
        """
        """
        self._record[property] = value
        return True

    
    

    
    def load(self, value):
        """
        """
        self._record = value
        return True


    def dump(self): 
        """
        """
        return copy.deepcopy(self._record)
        

    def set_json(self, value):
        """
        """
        record = json.loads(value)
        self.load(record)
        return True

    def get_json(self):
        """
        """
        return json.dumps(self.dump())

    @property
    def record(self):
        return self.dump()

    @record.setter
    def record(self, value):
        return self.load(value)
    
    @property
    def json(self):
        return self.get_json()

    @json.setter
    def json(self, value):
        return self.set_json(value)


    """
    User creation
    """

    def create(self):
        """Create a new user
        """
        self._record = m.create_user(self.record)
        return
    
    """
    API I/O
    """

    def api_get(self):
        """Retrieve user record from api
        """
        record = m.get_user(self.program, self.membershipNumber) 
        self._record = record if record else self._record
        self._in_db = True if record else False
        return True if record else False
    
    
    def api_post(self):
        """Post user record to api
        """
        self._in_db = True
        return m.post(self.record)

    def api_delete(self):
        """Delete user in api. 
        """
        self._in_db = False
        m.delete(self.record)
        self._record = m.get_base_record()
        return

    """
    Flask-login components
    """
    @property
    def is_authenticated(self):
        """
        """
        return m.is_logged_in(self.record)


    @property
    def is_active(self):
        """
        """
        return m.get_active_status(self.record)
        

    @property
    def is_anonymous(self):
        """
        """
        return False if self.id else True


    def get_id(self):
        """
        """
        return self._record.get('membershipNumber', None)



    """
    Properties
    """
    def set_email(self, value):
        """
        """
        self._record['membershipNumber'] = value
        self._record['member']['email'] = value
    

    @property
    def email(self):
        return self._record.get('member', {}).get('email', None)

    @email.setter
    def email(self, value):
        self.set_email(value)

    
    def set_programName(self, value):
        """
        """
        value = value.replace('http://', 'https://')
        self._record['programName'] = value

    def get_programName(self):
        """
        """
        return self._record.get('programName', None)
        
    
    @property
    def program(self):
        return self.get_programName()

    @program.setter
    def program(self, value):
        self.set_programName(value)

    def set_membershipNumber(self, value):
        """
        """
        self._record['membershipNumber'] = value

    @property
    def membershipNumber(self):
        return self._record.get('membershipNumber', None)

    @membershipNumber.setter
    def membershipNumber(self, value):
        self.set_membershipNumber(value)

    
    def get(self, userid):
        """
        """
        self._record = m.get_user(self.program, userid)
        return   

    """
    Password management
    """
    def add_credential(self, password_value):
        """
        """
        self.record = m.add_credential(self.record, password_value, self._salt)
        
    def set_password(self, password_value):
        """
        """
        self.record = m.add_credential(self.record, password_value, self._salt)

    def get_hash(self, password_value):
        """
        """
        return m.get_hash_password(password_value, self._salt)

    def authenticate(self, password):
        """Check if password matches hash on record
        """
        return m.authenticate_user(self.record, password, self._salt)

    """
    Role management
    """
    
    def set_active(self):
        self._record = m.set_active_status(self.record)

    def set_inactive(self):
        self._record = m.set_inactive_status(self.record)

    def add_role(self, role):
        self._record = m.add_role(self.record, role)

    def remove_role(self, role):
        self._record = m.remove_role(self.record, role)
    
    def get_active_roles(self):
        return m.get_active_roles(self.record)

    @property
    def roles(self):
        return self.get_active_roles()


    """
    Events
    """
    def login(self, remote_addr=None):
        self._record = m.add_login_event(self.record, remote_addr)
        self.api_post()

    def login_fail(self, remote_addr=None):
        self._record = m.add_failed_login_event(self.record, remote_addr)
        self.api_post()
    
    def logout(self, remote_addr=None):
        self._record = m.add_logout_event(self.record, remote_addr)
        self.api_post()
    


    """
    HTML templates
    """
    def get_login_form(self, referrer=None, action_url='/login'):
        """
        """
        return m.get_login_form(self.membershipNumber, referrer, action_url)

    def get_register_form(self, referrer=None, action_url='/register'):
        """
        """
        return m.get_register_form(self.membershipNumber, referrer, action_url)





from flask import Flask
from flask import request
from flask import Response
from flask import redirect
from flask import url_for
from flask import jsonify
from kraken_user.helpers import json

from kraken_user import kraken_user as m
from kraken_user.class_kraken_user import User
from kraken_user.class_kraken_users import Users

UPLOAD_FOLDER = '/path/to/the/uploads'

# Initialize flask app
app = Flask(__name__,
            static_url_path='',
            static_folder='static',
            template_folder='templates')
app.secret_key = b'_5#mn"F4Q8znxec]/'
app.config['SEND_FILE_MAX_AGE_DEFAULT'] = 0
app.config['UPLOAD_FOLDER'] = UPLOAD_FOLDER


@app.route('/', methods=['GET', 'POST'])
def main_get():

    key = 'name'
    value = request.args.get(key)

    if value:
        r = User()
        records = r.autocomplete(key, '%' + str(value) + '%')
        return jsonify(records)


    content = "Api for kraken_user"
    return Response(content)


@app.route('/<key>/<value>', methods=['GET', 'POST'])
def search_path_get(key, value):

    r = User()
    records = r.search(key, '%' + value + '%')
    return jsonify(records)


@app.route('/autocomplete', methods=['GET', 'POST'])
def autocomplete_params_get():

    key = 'name'
    value = request.args.get(key)

    r = User()
    records = r.autocomplete(key, '%' + str(value) + '%')
    return jsonify(records)



@app.route('/autocomplete/<key>/<value>', methods=['GET', 'POST'])
def autocomplete_path_get(key, value):

    r = User()
    records = r.autocomplete(key, '%' + value + '%')
    return jsonify(records)


@app.route('/log', methods=['POST'])
def log_post(key, value):
    """Registers a log event
    """
    
    return Response('')

@app.route('/about', methods=['GET'])
def about_get(key, value):
    """Returns instrument record for api
    """
    

    record = m.get_instrument()
    return jsonify(record)



def run_api():
    app.run(host='0.0.0.0', debug=False)


    